/*
 *	GENERA.C
 *	LaMoSca (LAboratorio di MOtori per SCAcchi)
 *	versione 0.05
 *	2001 Pietro Valocchi
 */


#include <stdio.h>
#include <string.h>
#include "Common.h"
#include "Extern.h"


/* stack delle mosse pseudolegali */
Mossa StackMosse[MAXMOSSE];

/* memoria mosse eseguite */
Casa History;

/* indice del livello di analisi */
int Livello;

/* posizione iniziale della scacchiera */
int ColoriIniziali[NUMEROCASE] = {
	NERO,   NERO,   NERO,   NERO,   NERO,   NERO,   NERO,   NERO,
	NERO,   NERO,   NERO,   NERO,   NERO,   NERO,   NERO,   NERO,
	VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,
	VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,
	VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,
	VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,
	BIANCO, BIANCO, BIANCO, BIANCO, BIANCO, BIANCO, BIANCO, BIANCO,
	BIANCO, BIANCO, BIANCO, BIANCO, BIANCO, BIANCO, BIANCO, BIANCO
};

int PezziIniziali[NUMEROCASE] = {
	TORRE,  CAVALLO, ALFIERE, REGINA, RE,     ALFIERE, CAVALLO, TORRE,
	PEDONE, PEDONE,  PEDONE,  PEDONE, PEDONE, PEDONE,  PEDONE,  PEDONE,
	VUOTO,  VUOTO,   VUOTO,   VUOTO,  VUOTO,  VUOTO,   VUOTO,   VUOTO,
	VUOTO,  VUOTO,   VUOTO,   VUOTO,  VUOTO,  VUOTO,   VUOTO,   VUOTO,
	VUOTO,  VUOTO,   VUOTO,   VUOTO,  VUOTO,  VUOTO,   VUOTO,   VUOTO,
	VUOTO,  VUOTO,   VUOTO,   VUOTO,  VUOTO,  VUOTO,   VUOTO,   VUOTO,
	PEDONE, PEDONE,  PEDONE,  PEDONE, PEDONE, PEDONE,  PEDONE,  PEDONE,
	TORRE,  CAVALLO, ALFIERE, REGINA, RE,     ALFIERE, CAVALLO, TORRE
};


/* Indica se un pezzo puo' muovere di piu' case alla volta
   0 PEDONE, 1 CAVALLO, 2 ALFIERE, 3 TORRE, 4 REGINA, 5 RE 
   Il pedone e' trattato separatamente */
BOOL Trascina[NUMEROPEZZI] = {
	FALSE, FALSE, TRUE, TRUE, TRUE, FALSE
};

/* Numero delle direzioni in cui un pezzo puo' muovere
   0 PEDONE, 1 CAVALLO, 2 ALFIERE, 3 TORRE, 4 REGINA, 5 RE 
   Il pedone e' trattato separatamente */
int NumeroDirezioni[NUMEROPEZZI] = {
	0, 8, 4, 4, 8, 8
};

/* Direzioni dei vari pezzi in una matrice 10*12 (vedi matrice Da10x12A8x8[])
   0 PEDONE, 1 CAVALLO, 2 ALFIERE, 3 TORRE, 4 REGINA, 5 RE
   Es.: cavallo -21 vuol dire 2 righe sotto ed una colonna a sinistra
   Il pedone e' trattato separatamente */
int OffsetPezzi[NUMEROPEZZI][8] = {
	{   0,   0,   0,  0, 0,  0,  0,  0 },
	{ -21, -19, -12, -8, 8, 12, 19, 21 },
	{ -11,  -9,   9, 11, 0,  0,  0,  0 },
	{ -10,  -1,   1, 10, 0,  0,  0,  0 },
	{ -11, -10,  -9, -1, 1,  9, 10, 11 },
	{ -11, -10,  -9, -1, 1,  9, 10, 11 }
};

/* Da12A8[] e' una matrice 10*12 utilizzata per verificare la legalita' 
   delle mosse. E' ottenuta aggiungendo ad una matrice 8*8 una 
   colonna per lato e due righe superiori ed inferiori con valori pari a -1,
   che semplificano il test di mosse che cadono al di fuori della scacchiera.
   Sono necessarie due righe per tener conto delle due case della mossa del cavallo,
   cosa che orizzontalmente e' fornita dell'unione delle colonne destra e sinistra.
   Le celle diverse da -1 rappresentano gli indici nella scacchiera 8*8 */
int Da12A8[120] = {
	 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	 -1,  0,  1,  2,  3,  4,  5,  6,  7, -1,
	 -1,  8,  9, 10, 11, 12, 13, 14, 15, -1,
	 -1, 16, 17, 18, 19, 20, 21, 22, 23, -1,
	 -1, 24, 25, 26, 27, 28, 29, 30, 31, -1,
	 -1, 32, 33, 34, 35, 36, 37, 38, 39, -1,
	 -1, 40, 41, 42, 43, 44, 45, 46, 47, -1,
	 -1, 48, 49, 50, 51, 52, 53, 54, 55, -1,
	 -1, 56, 57, 58, 59, 60, 61, 62, 63, -1,
	 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1
};

/* Per ogni casa della scacchiera 8*8, la matrice Da8x8A10x12[] fornisce 
   l'indice della corrispondente casa nella matrice 10*12 Da10x12A8x8[].
   L'utilizzo sara' del tipo (vedi la funzione GeneraMossePseudoLegali() )
   CasaDestinazione = Da12A8[Da8A12[CasaDestinazione] + OffsetPezzi[Pezzi[CasaCorrente]][Direzione]];
*/
int Da8A12[NUMEROCASE] = {
	21, 22, 23, 24, 25, 26, 27, 28,
	31, 32, 33, 34, 35, 36, 37, 38,
	41, 42, 43, 44, 45, 46, 47, 48,
	51, 52, 53, 54, 55, 56, 57, 58,
	61, 62, 63, 64, 65, 66, 67, 68,
	71, 72, 73, 74, 75, 76, 77, 78,
	81, 82, 83, 84, 85, 86, 87, 88,
	91, 92, 93, 94, 95, 96, 97, 98
};


/* SalvaMossa() inserisce una mossa nello stack */
void SalvaMossa(int Da, int A) {
	StackMosse[InizioMosse[Livello]].Da = (char)Da;
	StackMosse[InizioMosse[Livello]].A  = (char)A;
	InizioMosse[Livello]++;
}


/* GeneraPseudo() genera le mosse legali per la posizione corrente */
void GeneraPseudo() {
	int CasaCorrente, CasaDestinazione, Direzione;

	/* azzera il contatore delle mosse */
	InizioMosse[Livello] = InizioMosse[Livello-1];

	/* cerca tutte le case per pezzi del colore che deve muovere */
	for (CasaCorrente = 0; CasaCorrente < NUMEROCASE; CasaCorrente++) {
		if (B[CasaCorrente].Colore == ChiMuove) {
			if (B[CasaCorrente].Pezzo == PEDONE) {
				if (ChiMuove == BIANCO) {
					/* mosse per il pedone bianco */
					if (COLONNA(CasaCorrente) != 0 && B[CasaCorrente - 9].Colore == NERO)
						/* mangia in avanti a sinistra */
						SalvaMossa(CasaCorrente, CasaCorrente - 9);
					if (COLONNA(CasaCorrente) != 7 && B[CasaCorrente - 7].Colore == NERO)
						/* mangia in avanti a destra */
						SalvaMossa(CasaCorrente, CasaCorrente - 7);
					if (B[CasaCorrente - CASEPERRIGA].Colore == VUOTO) {
						/* avanti una casa */
						SalvaMossa(CasaCorrente, CasaCorrente - 8);
						if (CasaCorrente >= 48 && B[CasaCorrente - 16].Colore == VUOTO)
							/* avanti due case */
							SalvaMossa(CasaCorrente, CasaCorrente - 16);
					}
				}
				else {
					/* mosse per il pedone nero */
					if (COLONNA(CasaCorrente) != 0 && B[CasaCorrente + 7].Colore == BIANCO)
						SalvaMossa(CasaCorrente, CasaCorrente + 7);
					if (COLONNA(CasaCorrente) != 7 && B[CasaCorrente + 9].Colore == BIANCO)
						SalvaMossa(CasaCorrente, CasaCorrente + 9);
					if (B[CasaCorrente + CASEPERRIGA].Colore == VUOTO) {
						SalvaMossa(CasaCorrente, CasaCorrente + 8);
						if (CasaCorrente <= 15 && B[CasaCorrente + 16].Colore == VUOTO)
							SalvaMossa(CasaCorrente, CasaCorrente + 16);
					}
				}
			} else {
				/* per i pezzi differenti dal pedone */
				/* genera mosse per ciascuna delle direzioni legali */
				for (Direzione = 0; Direzione < NumeroDirezioni[B[CasaCorrente].Pezzo]; Direzione++) {
					/* "CasaDestinazione" == casa destinazione inizializzata con la corrente */
					CasaDestinazione = CasaCorrente;
					while (TRUE) {
						/* calcola la casa destinazione sommando l'offset
						per il pezzo corrente alla casa corrente.
						Si passa per la matrice 10*12 per verificare che
						la mossa non cada fuori dalla scacchiera */
						CasaDestinazione = Da12A8[Da8A12[CasaDestinazione] + 
							OffsetPezzi[B[CasaCorrente].Pezzo][Direzione]];
						/* mossa fuori dalla scacchiera */
						if (CasaDestinazione == -1)
							break;
						/* mossa su una casa con un pezzo avversario */
						if (B[CasaDestinazione].Colore != VUOTO) {
							if (B[CasaDestinazione].Colore != ChiMuove)
								SalvaMossa(CasaCorrente, 
									CasaDestinazione);
							break;
						}
						/* mossa su una casa vuota */
						SalvaMossa(CasaCorrente, CasaDestinazione);
						/* se il pezzo puo' muovere di piu' case, continua */
						if (!Trascina[B[CasaCorrente].Pezzo])
							break;
					}
				}
			}
		}
	}
}


/* Inizializza() inizializza la scacchiera */
void Inizializza() {
	int i;

	for (i = 0; i < NUMEROCASE; i++) {
		B[i].Colore = ColoriIniziali[i];
		B[i].Pezzo  = PezziIniziali[i];
	}
	ChiMuove = BIANCO;
	ColoreProgramma = VUOTO;
}


/* Muovi() riporta una mossa sulla scacchiera */
void Muovi(Mossa *M) {

	/* salva informazioni sulla casa destinazione 
	per tornare eventualmente indietro */
	History.Colore = B[M->A].Colore;
	History.Pezzo  = B[M->A].Pezzo;

	/* esegui la mossa */
	B[M->A].Colore  = B[M->Da].Colore;
	B[M->A].Pezzo   = B[M->Da].Pezzo;
	B[M->Da].Colore = VUOTO;
	B[M->Da].Pezzo  = VUOTO;

	/* cambia colore */
	ChiMuove ^= 1;

	/* incrementa livello */
	Livello++;
}


/* TornaIndietro() ripristina una mossa salvata */
void TornaIndietro(Mossa *M) {

	/* decrementa livello */
	Livello--;

	/* cambia colore */
	ChiMuove ^= 1;

	/* ripristina la vecchia mossa */
	B[M->Da].Colore = B[M->A].Colore;
	B[M->Da].Pezzo  = B[M->A].Pezzo;
	B[M->A].Colore  = History.Colore;
	B[M->A].Pezzo   = History.Pezzo;
}


/* MossaAvversario() esegue una mossa dell'avversario */
BOOL MossaAvversario(char * Input) {
	BOOL Trovato = FALSE;
	int i, Da, A, ch;
	Mossa M;

	/* calcola dai primi due caratteri l'indice della casa di partenza */
	Da = (Input[0] - 'a') + CASEPERRIGA * (CASEPERRIGA - (Input[1] - '0'));
	/* calcola dai caratteri rimanenti l'indice della casa di destinazione */
	A  = (Input[2] - 'a') + CASEPERRIGA * (CASEPERRIGA - (Input[3] - '0'));

	/* genera lo stack delle mosse per l'avversario salvando il numero mosse generate */
	Livello = 1;
	GeneraPseudo();

	/* controlla che la mossa sia legale cercandola nello stack */
	for (i = InizioMosse[Livello-1]; i < InizioMosse[Livello]; i++) {
		if ((StackMosse[i].Da == Da) && (StackMosse[i].A == A)) {
			Muovi(&StackMosse[i]);
			Trovato = TRUE;

			/* verifica se si tratta di una promozione */
			if ((B[A].Pezzo == PEDONE) && ((RIGA(A) == 0) || (RIGA(A) == 7))) {
				ch = Input[4];
				B[StackMosse[i].A].Pezzo = strchr(SimboliPezziNero, ch) - SimboliPezziNero;
			}

			/* se  la prima mossa dell'avversario, 
			   fa partire il programma */
			if (!ForceMode)
				ColoreProgramma = ChiMuove;
			break;

		}
	}

	/* controlla se si tratta di un arrocco */
	if (ChiMuove == BIANCO) {
		if (!strcmp(Input, "e1g1")) {
			M.Da = H1; M.A = F1;
			Muovi(&M);
			M.Da = E1; M.A = G1;
			Muovi(&M);
			Trovato = TRUE;
		}
		if (!strcmp(Input, "e1c1")) {
			M.Da = A1; M.A = D1;
			Muovi(&M);
			M.Da = E1; M.A = C1;
			Muovi(&M);
			Trovato = TRUE;
		}
	} else {
		if (!strcmp(Input, "e8g8")) {
			M.Da = H8; M.A = F8;
			Muovi(&M);
			M.Da = E8; M.A = G8;
			Muovi(&M);
			Trovato = TRUE;
		}
		if (!strcmp(Input, "e8c8")) {
			M.Da = A8; M.A = D8;
			Muovi(&M);
			M.Da = E8; M.A = C8;
			Muovi(&M);
			Trovato = TRUE;
		}
	}

	return Trovato;
}

/* InScacco() verifica una mossa e torna il valore */
BOOL InScacco(int *ValoreCorrente) {

	int j;

	/* genera mosse per l'avversario per vedere se c'e' uno scacco */
	GeneraPseudo();

	/* verifica che il re non sia in una casa destinazione */
	for (j = InizioMosse[Livello-1]; j < InizioMosse[Livello]; j++) {
		if (B[StackMosse[j].A].Pezzo == RE) {
			return FALSE;
		}
	}

	/* valuta posizione */
	*ValoreCorrente = Valuta();

	return TRUE;
}


/* MossaProgramma() genera una mossa per il computer */
BOOL MossaProgramma(char *CoordinateMossa) {

	int   i, ValorePosizione, ValoreCorrente;
	Mossa M;
	BOOL  Trovata = FALSE;

	/* genera mosse per il programma */
	Livello = 1;
	GeneraPseudo();

	/* prova tutte le mosse fino a trovare la migliore */
	ValorePosizione = -INFINITO;
	if (LogFile) fprintf(LogFile, "\nMosse/Valori: ");

	for (i = InizioMosse[Livello-1]; i < InizioMosse[Livello]; i++) {

		/* esegui la mossa */
		Muovi(&StackMosse[i]);

		if (InScacco(&ValoreCorrente)) {
			if (ValoreCorrente > ValorePosizione) {
				Trovata = TRUE;
				ValorePosizione = ValoreCorrente;
				/* salva la mossa scelta */
				M.A  = StackMosse[i].A;
				M.Da = StackMosse[i].Da;
			}
		}

		/* ripristina la mossa */
		TornaIndietro(&StackMosse[i]);

		if (LogFile)
			fprintf(LogFile, "%c%d%c%d=%d ", 
				COLONNA(StackMosse[i].Da)+ 'a', 8 - RIGA(StackMosse[i].Da), 
				COLONNA(StackMosse[i].A) + 'a', 8 - RIGA(StackMosse[i].A),
				ValoreCorrente);
	}

	if (Trovata) {
		/* esegui la mossa */
		Muovi(&M);

		/* crea la stringa con le coordinate della mossa corrente */
		sprintf(CoordinateMossa, "%c%d%c%d",
			COLONNA(M.Da)+ 'a', 8 - RIGA(M.Da), 
			COLONNA(M.A) + 'a', 8 - RIGA(M.A));
		}

	return Trovata;
}


